Render3D = {}
r3dmt = {}

sqrt = math.sqrt

function Render3D:new(sEx,sEy,sEz,sw,sh,sminvdist)
	return setmetatable({Ex=sEx,Ey=sEy,Ez=sEz,w=sw,h=sh,x=0,y=0,z=0,rx=0,ry=0,rz=0,
						 lastD=smaxvdist,minvdist=sminvdist,polys={},nx=0,ny=0,nz=0,
						 lx=1,ly=0,lz=0,r=255,g=255,b=255,a=255,matrixstack={}},r3dmt)
end

function Render3D:GetRoto(x,y,z)
	local rx,ry,rz
	--z rotation
	x = x + self.x
	y = y + self.y
	z = z + self.z
	rx = x*math.cos(-self.rz) - y*math.sin(-self.rz)
	ry = x*math.sin(-self.rz) + y*math.cos(-self.rz)
	x = rx
	y = ry
	--x rotation
	ry = y* math.cos(-self.rx) + z*math.sin(-self.rx)
	rz = y* math.sin(-self.rx) - z*math.cos(-self.rx)
	y = ry
	z = rz
	--y rotation
	rx = x* math.cos(-self.ry) - z*math.sin(-self.ry)
	rz = x* math.sin(-self.ry) + z*math.cos(-self.ry)
	z = rz
	x = rx
	return -x,-y,z
end

function Render3D:GetPoint(x,y,z)
	--get rotation
	--homogeneous coordinates w value
	local fw = (z)/self.Ez
	--2D projection value (-1,1]
	local bx = ((x-self.Ex)/fw+1)*self.w/2
	local by = self.h-((y-self.Ey)/fw+1)*self.h/2
	--actual screen coordinates
	return bx,by
end

function Render3D:triangle(mode,x1,y1,z1,x2,y2,z2,x3,y3,z3)
	--check if in bounds
	if z1+self.z<=self.minvdist or z2+self.z<=self.minvdist or z3+self.z<=self.minvdist then
		return
	end
	x1,y1,z1 = self:GetRoto(x1,y1,z1)
	x2,y2,z2 = self:GetRoto(x2,y2,z2)
	x3,y3,z3 = self:GetRoto(x3,y3,z3)
	table.insert(self.polys,{x1,y1,z1,x2,y2,z2,x3,y3,z3,m=mode,color={self.r,self.g,self.b,self.a},normal={self.nx,self.ny,self.nz}})
end

function Render3D:normal(x,y,z)
	self.nx = x
	self.ny = y
	self.nz = z
end

function Render3D:color(r,g,b,a)
	self.r = r
	self.g = g
	self.b = b
	self.a = a
end

function Render3D:clear()
	self.polys = {}
end

function Render3D:present()
	local r,g,b,a = love.graphics.getColor()
	CurrentRender3D = self
	table.sort(self.polys,PolyCmp)
	for index,poly in pairs(self.polys) do
	    local n = poly.normal
		local lnx,lny,lnz = self.lx,self.ly,self.lz
		local C = DotProduct(lnx,lny,lnz,n[1],n[2],n[3])
		love.graphics.setColor(poly.color[1]/2-C*poly.color[1]/2,poly.color[2]/2-C*poly.color[2]/2,poly.color[3]/2-C*poly.color[3]/2,poly.color[4])
		local bx1,by1,bx2,by2,by3,bx3 
		bx1,by1 = self:GetPoint(poly[1],poly[2],poly[3])
		bx2,by2 = self:GetPoint(poly[4],poly[5],poly[6])
		bx3,by3 = self:GetPoint(poly[7],poly[8],poly[9])
		love.graphics.triangle(poly.m,bx1,by1,bx2,by2,bx3,by3)
	end
	love.graphics.setColor(r,b,g,a)
end

function Render3D:translate(x,y,z)
	self.x = self.x + x
	self.y = self.y + y
	self.z = self.z + z
end

function Render3D:rotate(rx,ry,rz)
	self.rx = self.rx + rx
	self.ry = self.ry + ry
	self.rz = self.rz + rz
end

function Render3D:push()
	--store all infomation on a stack
	table.insert(self.matrixstack,{self.Ex,self.Ey,self.Ez,self.w,self.h,self.x,self.y,self.z,self.rx,self.ry,self.rz,
				 self.minvdist,self.nx,self.ny,self.nz,self.lx,self.ly,self.lz})
end

function Render3D:pop()
	local m = self.matrixstack[#self.matrixstack]
	table.remove(self.matrixstack)
	self.Ex = m[1]
	self.Ey = m[2]
	self.Ez = m[3]
	self.w = m[4]
	self.h = m[5]
	self.x = m[6]
	self.y = m[7]
	self.z = m[8]
	self.rx = m[9]
	self.ry = m[10]
	self.rz = m[11]
	self.minvdist = m[12]
	self.nx = m[13]
	self.ny = m[14]
	self.nz = m[15]
	self.lx = m[16]
	self.ly = m[17]
	self.lz = m[18]
end

r3dmt.__index = Render3D

function Max(a,b,c)
	if a>b then
		if a>c then 
			return a
		else
			return c
		end
	else
		if b>c then
			return b
		else
			return c
		end
	end
end

function Min(a,b,c)
	if a<b then
		if a<c then 
			return a
		else
			return c
		end
	else
		if b<c then
			return b
		else
			return c
		end
	end
end

function AbsMax(a,b,c)
	return math.max(math.abs(a),math.min(math.abs(b),math.abs(c)))
end

function PolyCmp(a,b)
	local a1,a2,a3 = Dist(a[1],a[2],a[3]),Dist(a[4],a[5],a[6]),Dist(a[7],a[8],a[9])
	local b1,b2,b3 = Dist(b[1],b[2],b[3]),Dist(b[4],b[5],b[6]),Dist(b[7],b[8],b[9])
	local am = Max(a1,a2,a3)
	local bm = Max(b1,b2,b3)
	if a~=b and am==bm then
		a1,a2,a3 = TriMiddle(a[1],a[2],a[3],a[4],a[5],a[6],a[7],a[8],a[9])
		b1,b2,b3 = TriMiddle(b[1],b[2],b[3],b[4],b[5],b[6],b[7],b[8],b[9])
		am = Dist(a1,a2,a3)
		bm = Dist(b1,b2,b3)
		local acm = Min(a[2],a[5],a[8])
		local bcm = Min(b[2],b[5],b[8])
		if acm<0 or bcm<0 then
			return am>bm
		end
		acm = Min(a[1],a[4],a[7])
		bcm = Min(b[1],b[4],b[7])
		if acm<0 or bcm<0 then
			return am>bm
		end
		return am<bm
	end
	return am>bm
end

function SubVec3(x1,y1,z1,x2,y2,z2)
	return x1-x2,y1-y2,z1-z2
end

function CrossProduct(x1,y1,z1,x2,y2,z2)
	local i = y1*z2-z1*y2
	local j = z1*x2-x1*z2
	local k = x1*y2-y1*x2
	return i,j,k
end

function Normal(x1,y1,z1,x2,y2,z2,x3,y3,z3)
	--get vectors
	local p0x,p0y,p0z = SubVec3(x1,y1,z1,x2,y2,z2)
	local p1x,p1y,p1z = SubVec3(x2,y2,z2,x3,y3,z3)
	--take there cross product
	local nx,ny,nz = CrossProduct(p0x,p0y,p0z,p1x,p1y,p1z)
	--normilize
	local n = math.sqrt(nx^2 + ny^2 + nz^2)
	return nx/n,ny/n,nz/n
end

function DotProduct(x1,y1,z1,x2,y2,z2)
	return x1*x2+y1*y2+z1*z2
end

function TriMiddle(x1,y1,z1,x2,y2,z2,x3,y3,z3)
	return (x1+x2+x3)/3,(y1+y2+y3)/3,(z1+z2+z3)/3
end

function Dist(x,y,z)
	return sqrt(x*x+y*y+z*z)
end